/*
 * Dummy project to validate header files
 *
 * This project is not intended to be executed, it should only include all
 * header files to make sure that they can be used with stricter compiler
 * settings than the libgit2 source files generally supports.
 */
#include "git2.h"

int main(void)
{
    return 0;
}
